"""filewritejson.py unit tests."""
import io
from unittest.mock import mock_open, patch

import pytest

from pypyr.context import Context
from pypyr.errors import (
    ContextError,
    KeyInContextHasNoValueError,
    KeyNotInContextError)
import pypyr.steps.filewritejson as filewrite


def test_filewritejson_no_filewritejson_raises():
    """No input fileWriteJson raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteJson'] doesn't exist. "
        "It must exist for pypyr.steps.filewritejson.")


def test_filewritejson_none_filewritejson_raises():
    """None fileWriteJson raises."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': None})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteJson'] must have a value for "
        "pypyr.steps.filewritejson.")


def test_filewritejson_filewritejson_not_iterable_raises():
    """Not iterable fileWriteJson raises."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': 1})

    with pytest.raises(ContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteJson'] must exist, be iterable "
        "and contain 'path' for pypyr.steps.filewritejson. argument of type "
        "'int' is not iterable")


def test_filewritejson_empty_path_raises():
    """Empty path raises."""
    context = Context({
        'fileWriteJson': {
            'path': None
        }})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteJson']['path'] must have a value for "
        "pypyr.steps.filewritejson.")


def test_filewritejson_no_path_raises():
    """No path raises."""
    context = Context({
        'fileWriteJson': 'blah',
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == ("context['fileWriteJson']['path'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.filewritejson.")


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_no_payload(mock_path):
    """Success case writes all context out when no payload."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': {
            'path': '/arb/blah'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 2, "context should have 2 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson'] == {'path': '/arb/blah'}

        mock_path.assert_called_once_with('/arb/blah')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)
        # json well formed & new lines and indents are where they should be
        assert out_text.getvalue() == ('{\n'
                                       '  "k1": "v1",\n'
                                       '  "fileWriteJson": {\n'
                                       '    "path": "/arb/blah"\n'
                                       '  }\n'
                                       '}')


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_with_payload(mock_path):
    """Success case writes only specific context payload."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': {
            'path': '/arb/blah',
            'payload': [
                'first',
                'second',
                {'a': 'b', 'c': 123.45, 'd': [0, 1, 2]},
                12,
                True
            ]
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 2, "context should have 2 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson']['payload'] == [
            'first',
            'second',
            {'a': 'b', 'c': 123.45,
             'd': [0, 1, 2]},
            12,
            True
        ]
        assert context['fileWriteJson'] == {'path': '/arb/blah',
                                            'payload': [
                                                'first',
                                                'second',
                                                {'a': 'b',
                                                 'c': 123.45,
                                                 'd': [0, 1, 2]},
                                                12,
                                                True
                                            ]}

        mock_path.assert_called_once_with('/arb/blah')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == ('[\n'
                                       '  "first",\n'
                                       '  "second",\n'
                                       '  {\n'
                                       '    "a": "b",\n'
                                       '    "c": 123.45,\n'
                                       '    "d": [\n'
                                       '      0,\n'
                                       '      1,\n'
                                       '      2\n'
                                       '    ]\n'
                                       '  },\n'
                                       '  12,\n'
                                       '  true\n'
                                       ']')


@patch('pypyr.config.config.default_encoding', new='utf-16')
@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_with_payload_encoding_from_config(mock_path):
    """Success case writes only specific context payload with encoding."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': {
            'path': '/arb/blah',
            'payload': [
                'first',
                'second',
                {'a': 'b', 'c': 123.45, 'd': [0, 1, 2]},
                12,
                True
            ]
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 2, "context should have 2 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson']['payload'] == [
            'first',
            'second',
            {'a': 'b', 'c': 123.45,
             'd': [0, 1, 2]},
            12,
            True
        ]
        assert context['fileWriteJson'] == {'path': '/arb/blah',
                                            'payload': [
                                                'first',
                                                'second',
                                                {'a': 'b',
                                                 'c': 123.45,
                                                 'd': [0, 1, 2]},
                                                12,
                                                True
                                            ]}

        mock_path.assert_called_once_with('/arb/blah')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(
            mocked_path, 'w', encoding='utf-16')

        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == ('[\n'
                                       '  "first",\n'
                                       '  "second",\n'
                                       '  {\n'
                                       '    "a": "b",\n'
                                       '    "c": 123.45,\n'
                                       '    "d": [\n'
                                       '      0,\n'
                                       '      1,\n'
                                       '      2\n'
                                       '    ]\n'
                                       '  },\n'
                                       '  12,\n'
                                       '  true\n'
                                       ']')


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_no_payload_substitutions(mock_path):
    """Success case writes all context out with substitutions."""
    context = Context({
        'k1': 'v1',
        'pathkey': '/arb/path',
        'parent': [0, 1, {'child': '{k1}'}],
        'nested': '{parent[2][child]}',
        'fileWriteJson': {
            'path': '{pathkey}'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 5, "context should have 5 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson'] == {'path': '{pathkey}'}

        mock_path.assert_called_once_with('/arb/path')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)
        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == ('{\n'
                                       '  "k1": "v1",\n'
                                       '  "pathkey": "/arb/path",\n'
                                       '  "parent": [\n'
                                       '    0,\n'
                                       '    1,\n'
                                       '    {\n'
                                       '      "child": "v1"\n'
                                       '    }\n'
                                       '  ],\n'
                                       '  "nested": "v1",\n'
                                       '  "fileWriteJson": {\n'
                                       '    "path": "/arb/path"\n'
                                       '  }\n'
                                       '}')


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_with_payload_substitutions(mock_path):
    """Success case writes only specified context with substitutions."""
    context = Context({
        'k1': 'v1',
        'intkey': 3,
        'pathkey': '/arb/path',
        'parent': [0,
                   1,
                   {'child': ['{k1}',
                              '{intkey}',
                              ['a', 'b', 'c']
                              ]}],
        'nested': '{parent[2][child]}',
        'fileWriteJson': {
            'path': '{pathkey}',
            'payload': '{parent[2]}'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 6, "context should have 6 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson'] == {'path': '{pathkey}',
                                            'payload': '{parent[2]}'}
        assert context['parent'] == [0,
                                     1,
                                     {'child': ['{k1}',
                                                '{intkey}',
                                                ['a', 'b', 'c']
                                                ]
                                      }
                                     ]

        mock_path.assert_called_once_with('/arb/path')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)
        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == (
            '{\n'
            '  "child": [\n'
            '    "v1",\n'
            '    3,\n'
            '    [\n'
            '      "a",\n'
            '      "b",\n'
            '      "c"\n'
            '    ]\n'
            '  ]\n'
            '}')


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_with_payload_substitutions_encoding(mock_path):
    """Success case writes only specified context with substitutions."""
    context = Context({
        'k1': 'v1',
        'intkey': 3,
        'pathkey': '/arb/path',
        'enc': 'utf-32',
        'parent': [0,
                   1,
                   {'child': ['{k1}',
                              '{intkey}',
                              ['a', 'b', 'c']
                              ]}],
        'nested': '{parent[2][child]}',
        'fileWriteJson': {
            'path': '{pathkey}',
            'payload': '{parent[2]}',
            'encoding': '{enc}'
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 7, "context should have 7 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson'] == {'path': '{pathkey}',
                                            'payload': '{parent[2]}',
                                            'encoding': '{enc}'}
        assert context['parent'] == [0,
                                     1,
                                     {'child': ['{k1}',
                                                '{intkey}',
                                                ['a', 'b', 'c']
                                                ]
                                      }
                                     ]

        mock_path.assert_called_once_with('/arb/path')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(
            mocked_path, 'w', encoding='utf-32')
        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == (
            '{\n'
            '  "child": [\n'
            '    "v1",\n'
            '    3,\n'
            '    [\n'
            '      "a",\n'
            '      "b",\n'
            '      "c"\n'
            '    ]\n'
            '  ]\n'
            '}')


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_with_empty_payload(mock_path):
    """Empty payload write empty file."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': {
            'path': '/arb/blah',
            'payload': ''
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 2, "context should have 2 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson']['path'] == '/arb/blah'
        assert context['fileWriteJson']['payload'] == ''

        mock_path.assert_called_once_with('/arb/blah')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == '""'


@patch('pypyr.steps.filewritejson.Path')
def test_filewritejson_pass_with_none_payload(mock_path):
    """None payload write empty file."""
    context = Context({
        'k1': 'v1',
        'fileWriteJson': {
            'path': '/arb/blah',
            'payload': None
        }})

    with io.StringIO() as out_text:
        with patch('pypyr.steps.filewritejson.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_text.write
            filewrite.run_step(context)

        assert context, "context shouldn't be None"
        assert len(context) == 2, "context should have 2 items"
        assert context['k1'] == 'v1'
        assert context['fileWriteJson']['path'] == '/arb/blah'
        assert context['fileWriteJson']['payload'] is None

        mock_path.assert_called_once_with('/arb/blah')
        mocked_path = mock_path.return_value
        mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                         exist_ok=True)
        mock_output.assert_called_once_with(mocked_path, 'w', encoding=None)

        # json well formed & new lines + indents are where they should be
        assert out_text.getvalue() == "null"



"""An instance of a running Pipeline.

This contains the run-time information of a pipeline and a reference
to the pipeline body itself (the PipelineDefinition).

Don't confuse this with PipelineDefinition. PipelineDefinition is the yaml
body & loader properties. The Pipeline is specific to a single running instance
of a pipeline, whereas the PipelineDefinition is shared by different Pipeline
instances.
"""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
import copy
import logging
from os import PathLike
from pathlib import Path

from pypyr.cache.loadercache import loader_cache
from pypyr.cache.parsercache import contextparser_cache
from pypyr.config import config
from pypyr.context import Context
from pypyr.errors import Stop, StopPipeline, StopStepGroup
import pypyr.moduleloader
from pypyr.stepsrunner import StepsRunner

logger = logging.getLogger(__name__)


class Pipeline():
    """An instance of a running Pipeline.

    If you want to get pipeline properties from config.shortcuts if
    pipeline name matches a shortcut name, instantiate with the static factory
    Pipeline.new_pipe_and_args(). If you're not interested in
    config.shortcuts, you can instantiate with the default constructor instead.

    Use run() to run a pipeline. Use load_and_run_pipeline() to run a child
    pipeline from within an already running pipeline.

    If you specify py_dir, will add it to sys.path if it is not there already.

    Don't confuse a Pipeline with a PipelineDefinition. The Pipeline is the
    run-time properties of a single instance of a running pipeline, which
    includes the PipelineDefinition as a property. The PipelineDefinition is
    the pipeline yaml payload itself. A loader returns the PipelineDefinition.
    The PipelineDefinition is a globally shared cache of the pipeline body &
    meta-data.

    Attributes:
        name (str): Name of pipeline, sans .yaml at end.
        context_args (list[str]): All the input arguments after the pipeline
            name from cli.
        parse_input (bool): Default True. Run context_parser in pipeline.
        loader (str): Absolute name of pipeline loader module. If not specified
            will use pypyr.loaders.file.
        groups (list[str]): Step-group names to run in pipeline.
            Default if not set is ['steps'].
        success_group (str): Step-group name to run on success completion.
            Default if not set is on_success.
        failure_group (str: Step-group name to run on pipeline failure.
            Default if not set is on_failure.
        py_dir (Path-like): Custom python modules resolve from this dir.
        pipeline_definition (pypyr.pipedef.PipelineDefinition): The pipeline
            definition (its body/yaml payload) and loader information. Set by
            run(), not init.
        steps_runner (pypyr.stepsrunner.StepsRunner): StepsRunner instance that
            will run this pipeline's step-groups. Set by
            load_and_run_pipeline(), not init.
    """

    __slots__ = ['name', 'context_args', 'parse_input', 'loader', 'groups',
                 'success_group', 'failure_group', 'py_dir',
                 'pipeline_definition', 'steps_runner']

    # region constructors
    def __init__(self,
                 name: str,
                 context_args: list[str] | None = None,
                 parse_input: bool | None = True,
                 loader: str | None = None,
                 groups: list[str] | None = None,
                 success_group: str | None = None,
                 failure_group: str | None = None,
                 py_dir: str | bytes | PathLike | None = None) -> None:
        """Initialize a Pipeline.

        Args:
            name (str): Name of pipeline, sans .yaml at end.
            context_args (list[str]): All the input arguments after the
                pipeline name from cli.
            parse_input (bool): Default True. Run context_parser in pipeline.
            loader (str): Absolute name of pipeline loader module.
                        If not specified will use pypyr.loaders.file.
            groups (list[str]): Step-group names to run in pipeline.
                                Default if not set is ['steps'].
            success_group (str): Step-group name to run on success completion.
                                Default if not set is on_success.
            failure_group (str: Step-group name to run on pipeline failure.
                                Default if not set is on_failure.
            py_dir (Path-like): Custom python modules resolve from this dir.

        Returns:
            None
        """
        self.name = name
        self.context_args = context_args
        self.parse_input = parse_input
        self.loader = loader
        self.groups = groups
        self.success_group = success_group
        self.failure_group = failure_group
        self.py_dir = py_dir

        # initialize here, but use later
        # not using a classmethod fromLoader factory style thing coz PipeDef
        # AND StepsRunner depend on having a context object, which is subject
        # to logic only called later in obj life-time in load_and_run_pipeline.
        self.pipeline_definition = None
        self.steps_runner = None

    @classmethod
    def new_pipe_and_args(
        cls,
        name: str,
        context_args: list[str] | None = None,
        parse_input: bool | None = None,
        dict_in: dict | None = None,
        loader: str | None = None,
        groups: list[str] | None = None,
        success_group: str | None = None,
        failure_group: str | None = None,
        py_dir: str | bytes | PathLike | None = None) -> tuple[Pipeline,
                                                               dict | None]:
        """Return new Pipeline instance and dict_in args.

        Will initialize from config.shortcuts if arg `name` matches a shortcut.

        If shortcut doesn't exist fallback to the input args passed to new to
        initialize the Pipeline. If the shortcut does exist, any properties
        not specified by the shortcut configuration will fall back to this
        method's provided input arguments.

        When shortcut found, context_args will append to the end of
        `parser_args` as defined by shortcut. dict_in will merge into `args`
        as defined by shortcut.

        Returned Pipeline instance will default parse_input to True if
        shortcut.parser_args + context_args exist. If no context_args exist,
        but dict_in/shortcut.args do, returned Pipeline instance assumes
        parse_input=False if not explicitly set.

        Args:
            name (str): Name of pipeline, sans .yaml at end.
            context_args (list[str]): All the input arguments after the
                                      pipeline name from cli.
            parse_input (bool): Default True. Run context_parser in pipeline.
            dict_in (dict): Dict-like object to initialize the Context.
            loader (str): Absolute name of pipeline loader module.
                        If not specified will use pypyr.loaders.file.
            groups (list[str]): Step-group names to run in pipeline.
                                Default if not set is ['steps'].
            success_group (str): Step-group name to run on success completion.
                                Default if not set is on_success.
            failure_group (str: Step-group name to run on pipeline failure.
                                Default if not set is on_failure.
            py_dir (Path-like): Custom python modules resolve from this dir.

        Returns:
            New Pipeline instance, intialized from shortcut if name matches.
            Dict instance to use to initialize Context. This dict is the
            result of dict_in + input args found in shortcut, if any.
        """
        logger.debug("starting")
        if config.shortcuts:
            # assuming shortcuts is mostly empty dict, much faster to do truthy
            # if check before .get(), even though it looks redundant.
            shortcut = config.shortcuts.get(name)
            if shortcut:
                shortcut_name = name
                logger.debug("found shortcut in config for %s", shortcut_name)
                name = shortcut.get('pipeline_name')
                if not name:
                    raise pypyr.errors.ConfigError(
                        f"shortcut '{shortcut_name}' has no pipeline_name "
                        "set. You must set pipeline_name for this shortcut "
                        "in config so that pypyr knows which pipeline to run.")
                parser_args = shortcut.get('parser_args')
                if parser_args:
                    if isinstance(parser_args, str):
                        raise pypyr.errors.ConfigError(
                            f"shortcut '{shortcut_name}' parser_args should "
                            "be a list, not a string.")
                    # append context_args to shortcut's parser_args
                    context_args = (
                        parser_args + context_args if context_args
                        else parser_args)

                skip_parse = shortcut.get('skip_parse')
                # flip the bit - skip_parse means inverse of parse_args, but
                # only if it exists
                # since cli always set parse_args=True, the shortcut has
                # entirely to ignore the cli input to allow shortcut to calc
                # if parse_input should True/False based on args and
                # parser_args availability.
                parse_input = (
                    None if skip_parse is None else not skip_parse)

                shortcut_dict_in = shortcut.get('args')
                if shortcut_dict_in:
                    # deepcopy so downstream mutations don't touch the original
                    # dict in config
                    sc_dict = copy.deepcopy(shortcut_dict_in)
                    if dict_in:
                        # merge dict_in into shortcut's args
                        sc_dict.update(dict_in)
                    dict_in = sc_dict

                sc_groups = shortcut.get('groups', groups)
                # if config specified a str, take it to mean a single group
                groups = [sc_groups] if isinstance(sc_groups,
                                                   str) else sc_groups

                success_group = shortcut.get('success', success_group)
                failure_group = shortcut.get('failure', failure_group)
                loader = shortcut.get('loader', loader)
                dir_str = shortcut.get('py_dir')
                if dir_str:
                    # will still honor cwd for cli if not set, since it'll only
                    # override input if shortcut dir actually set.
                    py_dir = Path(dir_str)
            else:
                logger.debug("no shortcut found in config for %s", name)

        # if context_args exist, assume caller meant parse_input = True
        #  if no context_args but dict_in does exist, assume parse_input=False
        parse_input = cls._get_parse_input(parse_args=parse_input,
                                           args_in=context_args,
                                           dict_in=dict_in)
        pipeline = cls(name=name,
                       context_args=context_args,
                       parse_input=parse_input,
                       loader=loader,
                       groups=groups,
                       success_group=success_group,
                       failure_group=failure_group,
                       py_dir=py_dir)

        logger.debug("done")
        return pipeline, dict_in

    # endregion constructors

    def run(self, context: Context) -> None:
        """Run the current pypyr pipeline.

        Use me to run this pipeline.

        If you want to have the context available to you after the pipeline
        finishes, you (obviously) should provide an instance to the context
        parameter.

        Any exceptions raised from here indicate abnormal termination of a
        pipeline.

        Args:
            context (pypyr.context.Context): Any mutations of the context by
                the pipeline will be against this instance of it. If None, will
                create fresh new context with context_args using the pipeline's
                context_parser.

        Returns:
            None.
        """
        logger.debug("starting")

        try:
            self.load_and_run_pipeline(context)
        except Stop:
            logger.debug("Stop: stopped pypyr")

        logger.debug("done")

    def load_and_run_pipeline(self, context, parent=None):
        """Load and run the specified pypyr pipeline.

        Only use this from within an already running pipeline when it's
        calling a child pipeline.

        If you are running another pipeline from within a pipeline, call this,
        not run(). Do call run() instead for your 1st (root) pipeline if there
        are pipelines calling pipelines. The pypyr.steps.pype step uses
        load_and_run_pipeline to let pipelines call other pipelines.

        If you specify py_dir, will add it to sys.path if it is not there
        already.

        Args:
            context (pypyr.context.Context): Any mutations of the context by
                the pipeline will be against this instance of it. If None, will
                create fresh new context with context_args using the pipeline's
                context_parser.
            parent (Any): Passed to the loader. Can be anything of interest to
                the loader. In the case of the default file loader, this is
                a Path to the calling pipeline's parent directory.

        Returns:
            None.
        """
        logger.debug("you asked to run pipeline: %s", self.name)

        logger.debug("you set the initial context arg to: %s",
                     self.context_args)

        if context is None:
            context = Context()

        # add python module dir to sys.path before looking for loader
        if self.py_dir:
            pypyr.moduleloader.add_sys_path(self.py_dir)

        # could save loader_instance to self for >1 run on same pipeline, but
        # since you'd need extra check if self.loader has changed since last
        # time, O(1) dict lookup in cache prob not going to add too much
        # overhead by comparison.
        loader_instance = loader_cache.get_pype_loader(self.loader)

        # pipeline loading deliberately outside try catch. If the pipeline
        # doesn't exist there is no failure handler that can possibly run so
        # this is very much a fatal stop error.
        self.pipeline_definition = loader_instance.get_pipeline(name=self.name,
                                                                parent=parent)

        # add current pipeline's info to the callstack & remove when pipeline
        # done.
        with context.pipeline_scope(self):
            self._run_pipeline(context)

    def _run_pipeline(self, context):
        """Execute the internal implementation of the logic to run a pipeline.

        Don't call me unless you really know what you're doing - this method
        relies on setup done in the public method run() and
        load_and_run_pipeline().

        At this point the loader has already found and cached the
        PipelineDefinition, so this method is the actual runtime logic on the
        pipeline payload.

        context *must* be inside a pipeline_scope for this method to work
        as intended. pipeline_definition must be set on the current class
        instance.

        Args:
            context (pypyr.context.Context): Any mutations of the context by
                the pipeline will be against this instance of it. If None, will
                create fresh new context with context_args using the pipeline's
                context_parser.

        Returns:
            None.
        """
        logger.debug("starting")

        groups = self.groups
        success_group = self.success_group
        failure_group = self.failure_group

        if not groups:
            groups = [config.default_group]

            if not self.success_group and not self.failure_group:
                success_group = config.default_success_group
                failure_group = config.default_failure_group

        steps_runner = StepsRunner(
            pipeline_body=self.pipeline_definition.pipeline,
            context=context)

        self.steps_runner = steps_runner

        try:
            self._prepare_context(context)
        except Exception:
            # yes, yes, don't catch Exception. Have to, though, to run failure
            # handler. Also, it does raise it back up.
            logger.error("Something went wrong. Will now try to run %s",
                         failure_group)

            # failure_step_group will log but swallow any errors except Stop
            try:
                steps_runner.run_failure_step_group(failure_group)
            except StopStepGroup:
                pass

            logger.debug("Raising original exception to caller.")
            raise

        try:
            steps_runner.run_step_groups(groups=groups,
                                         success_group=success_group,
                                         failure_group=failure_group)
        except StopPipeline:
            logger.debug("StopPipeline: stopped %s", self.name)

        logger.debug("done")

    # region cli context input args
    def _prepare_context(self, context):
        """Prepare context for pipeline run.

        Args:
            context (pypyr.context.Context): Merge any new context generated
                from self.context_args by the pipeline's context_parser into
                this Context instance.

        Returns:
            None. The context instance passed in mutates, it's not passed back
            as a method return.
        """
        logger.debug("starting")

        if self.parse_input:
            logger.debug("executing context_parser")
            parsed_context = self._get_parsed_context()

            if parsed_context:
                context.update(parsed_context)
        else:
            logger.debug("skipping context_parser")

        logger.debug("done")

    def _get_parsed_context(self):
        """Execute get_parsed_context handler if specified.

        Dynamically load the module specified by the context_parser key in the
        pipeline dict and execute the get_parsed_context function on that
        module.

        Returns:
            dict-like instance, or None.

        Raises:
            AttributeError: parser specified on pipeline missing
                            get_parsed_context function.
        """
        logger.debug("starting")

        pipeline_yaml = self.pipeline_definition.pipeline
        parser_module_name = pipeline_yaml.get('context_parser')

        if parser_module_name:
            logger.debug("context parser specified: %s", parser_module_name)
            get_parsed_context = contextparser_cache.get_context_parser(
                parser_module_name)

            logger.debug("running parser %s", parser_module_name)

            result_dict = get_parsed_context(self.context_args)

            logger.debug("context parse %s done", parser_module_name)

            if result_dict is None:
                logger.debug(
                    "%s returned None. Using empty context instead",
                    parser_module_name
                )

            logger.debug("done")
            return result_dict
        else:
            logger.debug("pipeline does not have a context parser specified. "
                         "Initializing with empty context.")
            logger.debug("done")
            return None

    @staticmethod
    def _get_parse_input(parse_args, args_in, dict_in):
        """Return default for parse_input.

        This is to decide if context_parser should run or not.

        To make it easy on an API consumer, default behavior is ALWAYS to run
        parser UNLESS dict_in initializes context and there is no args_in.

        If dict_in specified, but no args_in: False
        If dict_in specified, AND args_in too: True
        If no dict_in specified, but args_in is: True
        If no dict_in AND no args_in: True
        If parse_args explicitly set, always honor its value.

        Args:
            parse_args (bool): Whether to run context parser.
            args_in (list[str]): String arguments as passed from the cli.
            dict_in (dict): Initialize context with this dict.

        Returns:
            Boolean. True if should parse input.
        """
        if parse_args is None:
            return not (not args_in and dict_in is not None)
        return parse_args
    # endregion cli context input args



"""pypyr step that adds items to a set."""
import logging

from pypyr.utils.asserts import assert_key_exists, assert_key_is_truthy

logger = logging.getLogger(__name__)


def run_step(context):
    """Add item(s) to a set.

    Expects input:
        add:
            set (set or str): Add addMe to this set.
            addMe (any): Add this item to set.
            unpack (bool): Optional. Defaults False. If True, enumerate addMe
                and add each item individually.

    If add.set is a str, it refers to a key in context which contains a
    set, e.g context['my_set'] = {1, 2, 3}. If no such key exists, will create
    a set with that name and add addMe as the 1st item on the new set.

    If you want to add/append to a list, use pypyr.steps.append instead.

    If addMe is enumerable and you want to add its children individually, set
    add.unpack = True.

    Args:
        context (pypyr.context.Context): Mandatory. Context is dict-like.
            Context must contain key 'add'
    """
    logger.debug("started")

    context.assert_key_has_value(key='add', caller=__name__)

    step_input = context.get_formatted('add')

    assert_key_is_truthy(obj=step_input,
                         key='set',
                         caller=__name__,
                         parent='add')

    assert_key_exists(obj=step_input,
                      key='addMe',
                      caller=__name__,
                      parent='add')

    the_set = step_input['set']
    add_me = step_input['addMe']
    is_update = step_input.get('unpack', None)

    # default to unpack/update for lists specifically
    # no real reason other than sort of feels like expected
    # behavior from a pipeline perspective.
    if is_update is None:
        is_update = isinstance(add_me, list)

    # str value means referring to a key in context rather than set instance
    if isinstance(the_set, str):
        existing_set = context.get(the_set, None)

        if existing_set:
            add_or_update_set(existing_set, add_me, is_update)
        else:
            # {x} works only f x single hashable, set(x) when x iterable
            context[the_set] = set(add_me) if is_update else {add_me}
    else:
        add_or_update_set(the_set, add_me, is_update)

    logger.debug("done")


def add_or_update_set(the_set, add_me, is_update):
    """Add or update the set.

    Args:
        the_set (set): The set to add/update
        add_me (any): Item(s) to add to the_set
        is_update (bool): If True does update rather than add.

    Returns: None
    """
    if is_update:
        the_set.update(add_me)
    else:
        the_set.add(add_me)



"""pyimport.py unit tests."""
import builtins
from pathlib import Path

from pypyr.context import Context
from pypyr.dsl import PyString
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.pyimport as pyimport
import pytest


def test_pyimport_none():
    """No pyimport input raises."""
    with pytest.raises(KeyNotInContextError) as err_info:
        context = Context({'blah': 'blah blah'})
        pyimport.run_step(context)

    assert str(err_info.value) == ("context['pyImport'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.pyimport.")


def test_pyimport_key_exists_but_none():
    """None source string raises."""
    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        context = Context({'pyImport': None})
        pyimport.run_step(context)

    assert str(err_info.value) == (
        "context['pyImport'] must have a value for pypyr.steps.pyimport.")


def test_pyimport_empty():
    """Empty source string imports nothing."""
    context = Context({'pyImport': ''})
    pyimport.run_step(context)
    assert context._pystring_globals == {}
    # only builtins
    assert list(dict.items(context._pystring_namespace)) == [
        ('__builtins__', builtins.__dict__)]


def test_pyimport():
    """Import namespace saved to context."""
    source = """\
import math

import tests.arbpack.arbmod
import tests.arbpack.arbmod2

from tests.arbpack.arbmultiattr import arb_attr, arb_func as y
"""
    context = Context({'pyImport': source})
    pyimport.run_step(context)
    ns = context._pystring_globals
    len(ns) == 4

    assert ns['math'].sqrt(4) == 2
    # no return value but shouldn't raise not found.
    ns['tests'].arbpack.arbmod.arbmod_attribute()
    assert ns['arb_attr'] == 123.456
    assert ns['y']('ab3') == 'ab3'

    # parent did not import anything NOT specified.
    # tests.arbpack.arbstep exists but wasn't specified for import.
    assert not hasattr(ns['tests'].arbpack, 'arbstep')
    # python has a quirk where if a submodule was imported anywhere else, it's
    # imported & cached for the global namespace of the parent package. So
    # arbmod3, which is used elsewhere in unit tests, might well be an
    # attribute of arbpack here even though not explicitly specified, if part
    # of the test-run included any code somewhere doing
    # import tests.arbpack.arbmod3.
    #
    # importlib.invalidate_caches() doesn't help
    # assert not hasattr(ns['tests'].arbpack, 'arbmod3') --> will FAIL


def test_pyimport_update_not_replace():
    """Import updates/merges to namespace dict, not overwrite."""
    context = Context({'pyImport': 'import math; import tests.arbpack.arbmod'})
    pyimport.run_step(context)

    context['pyImport'] = """\
from pathlib import Path
import tests.arbpack.arbmod3
"""

    # 2nd run should merge into ns dict, not overwrite.
    pyimport.run_step(context)
    ns = context._pystring_globals
    len(ns) == 3

    assert ns['math'].sqrt(4) == 2
    # both the arbpack imports valid because of python submodule global imports
    ns['tests'].arbpack.arbmod.arbmod_attribute()
    assert ns['tests'].arbpack.arbmod3.arb_func_in_arbmod3('ab3') == 'ab3'
    assert ns['Path'].cwd() == Path.cwd()


def test_pyimport_with_pystring():
    """Pystring evals with the globals namespace from pyimport."""
    context = Context({'a': -1,
                       'b': 'xx',
                       'c': 4,
                       'pyImport': 'import math',
                       'out': PyString('abs(a) + len(b) + math.sqrt(c)')})
    pyimport.run_step(context)

    assert context.get_formatted('out') == 5



"""Load pipelines from the filesystem."""
import logging
from pathlib import Path

from pypyr.cache.filecache import file_cache
from pypyr.config import config
from pypyr.errors import PipelineNotFoundError
from pypyr.moduleloader import add_sys_path
from pypyr.pipedef import PipelineDefinition, PipelineFileInfo
import pypyr.yaml

logger = logging.getLogger(__name__)

cwd_pipelines_dir = config.cwd.joinpath(config.pipelines_subdir)
pypyr_dir = Path(__file__).parents[1]
builtin_pipelines_dir = pypyr_dir.joinpath('pipelines')


# region find pipeline path
def find_pipeline(file_name, dirs):
    """Look for file_name in dirs.

    Args:
        file_name (str): String, likely {pipeline_name}.yaml
        dirs (list[(Path, str)]): List of (dir, msg). Msg prints
            to debug output if not found in this path.

    Returns:
        Resolved Path instance if found.

    Raises:
        PipelineNotFoundError: file_name not found in any of the dirs.
    """
    for parent in dirs:
        parent_dir = parent[0]
        path = parent_dir.joinpath(file_name)

        if path.is_file():
            logger.debug("Found %s", path)
            break
        else:
            logger.debug(parent[1], file_name)
    else:
        # didn't hit a break means no pipeline found anywhere
        searched_locations = "\n".join([str(p[0]) for p in dirs])
        raise PipelineNotFoundError(
            f"{file_name} not found in any of the following:\n"
            f"{searched_locations}")

    return path.resolve()


def get_pipeline_path(pipeline_name, parent):
    """Look for the pipeline in the various places it could be.

    1. If absolute path, just check that
    2. Check in parent, if it's specified
    3. Checks the cwd, if it's not the same as parent (2)
    4. Then check cwd/pipelines
    5. Then check {pypyr install dir}/pipelines dir.

    Args:
        pipeline_name (str): Name of pipeline to find
        parent (Path): 1st Path in which to look for pipeline_name.yaml

    Returns:
        Absolute Path to the pipeline_name.yaml file

    Raises:
        PipelineNotFoundError: if pipeline_name.yaml not anywhere.
    """
    logger.debug("starting")

    logger.debug("current parent is %s", parent)

    search_locations = []
    file_name = f'{pipeline_name}.yaml'

    # 1. absolute paths
    abs_candidate = Path(file_name)
    if abs_candidate.is_absolute():
        if abs_candidate.is_file():
            logger.debug("Found %s", abs_candidate)
            logger.debug("done")
            return abs_candidate.resolve()
        else:
            raise PipelineNotFoundError(f"{abs_candidate} does not exist.")
    else:
        # 2. parent/{pipeline_name}.yaml: go to 2 if parent == cwd
        if parent:
            parent = parent if isinstance(parent, Path) else Path(parent)
            # do a resolve so that full path in searched_locations err msg.
            parent = parent.resolve()
            # samefile raises err if either path doesn't .exist
            if parent.exists():
                if not parent.samefile(config.cwd):
                    search_locations.append((
                        parent,
                        "%s not found in parent pipeline directory. "
                        "Looking in cwd instead."))
            else:
                logger.debug(
                    'parent dir %s does not exist. skipping to cwd look-up.',
                    parent)

        # 3. {cwd}/{pipeline_name}.yaml
        search_locations.append((config.cwd,
                                "%s not found in cwd. "
                                 "Looking in 'cwd/pipelines' instead."))

        # 4. {cwd}/pipelines/{pipeline_name}.yaml
        search_locations.append((cwd_pipelines_dir,
                                "%s not found in cwd/pipelines. "
                                 "Looking in pypyr install directory instead.")
                                )

        # 5. {pypyr dir}/pipelines/{pipeline_name}.yaml
        search_locations.append((builtin_pipelines_dir,
                                "%s not found in {pypyr-dir}/pipelines."))

        pipeline_path = find_pipeline(file_name, search_locations)
        logger.debug("done")
        return pipeline_path

# endregion find pipeline path


def get_pipeline_definition(pipeline_name, parent):
    """Open and parse the pipeline definition yaml.

    Finds the yaml file, parses the yaml and returns a PipelineDefinition.

    pipeline_name.yaml should be in the parent/ directory, or in the
    fileloader directory look-up sequence.

    Args:
        pipeline_name (str): Name of pipeline. This will be the file-name of
                             the pipeline - i.e {pipeline_name}.yaml
        parent (Path-like): Start looking in parent/pipeline_name.yaml

    Returns:
        PipelineDefinition describing the pipeline. The dict parsed from the
            pipeline yaml is in its .pipeline property.
    """
    logger.debug("starting")

    pipeline_path = get_pipeline_path(pipeline_name=pipeline_name,
                                      parent=parent)

    logger.debug("Trying to open pipeline at path %s", pipeline_path)

    pipeline_definition = file_cache.get(
        str(pipeline_path),
        lambda: load_pipeline_from_file(pipeline_path))

    logger.debug("found %d stages in pipeline.",
                 len(pipeline_definition.pipeline))

    logger.debug("done")
    return pipeline_definition


def load_pipeline_from_file(path):
    """Load pipeline yaml from path on disk.

    Initialize a PipelineFileInfo for the pipeline with file loader specific
    properties.

    Args:
        path (Path-like): path to pipeline

    Returns:
        PipelineDefinition describing the pipeline, parsed from the yaml file
            at path.

    Raises:
        FileNotFoundError: path not found.
    """
    logger.debug("starting")

    try:
        with open(path, encoding=config.default_encoding) as yaml_file:
            pipeline_yaml = pypyr.yaml.get_pipeline_yaml(yaml_file)
    except FileNotFoundError:
        # this can only happen if file disappears between get_pipeline_path
        # & here, so pretty edge
        logger.error(
            "Couldn't open the pipeline. Looking for a file here: %s", path)
        raise

    # since path itself resolved, parent also already resolved.
    parent_dir = path.parent

    add_sys_path(parent_dir)

    info = PipelineFileInfo(pipeline_name=path.name,
                            parent=parent_dir,
                            loader=__name__,
                            path=path)

    pipeline_definition = PipelineDefinition(pipeline=pipeline_yaml,
                                             info=info)

    logger.debug("done")
    return pipeline_definition



"""Logging functions for pypyr.

Configuration for the python logging library.
"""
import logging
import logging.config
import sys

from pypyr.config import config

NOTIFY = 25


class LevelFilter(logging.Filter):
    """Only let through messages < level."""

    def __init__(self, level: int) -> None:
        """Set the level at which and lower this filter will log."""
        self.level = level

    def filter(self, record):
        """Pass through records with level less than configured level."""
        return record.levelno < self.level


def set_logging_config(log_level, handlers):
    """Set python logging library config.

    Run this ONCE at the start of your process. It formats the python logging
    module's output.
    """
    if log_level is None:
        level = NOTIFY
        format_string = config.log_notify_format
    else:
        level = log_level
        format_string = config.log_detail_format

    logging.basicConfig(
        format=format_string,
        datefmt=config.log_date_format,
        level=level,
        handlers=handlers)


def notify(self, msg, *args, **kwargs):
    """Log a message with severity 'NOTIFY' on the root logger."""
    if self.isEnabledFor(NOTIFY):
        self._log(NOTIFY, msg, args, **kwargs)


def set_up_notify_log_level():
    """Add up a global notify severity to the python logging package.

    NOTIFY severity is logging level between INFO and WARNING. By default it
    outputs only echo step and step name with description.

    This function should run once and only once at the initialization of pypyr.
    You shouldn't need to do so yourself, it's called from package init.
    """
    # could (should?) be checking hasattr like so:
    # hasattr(logging, levelName):
    # hasattr(logging, methodName):
    # hasattr(logging.getLoggerClass(), methodName):
    # but this extra check is arguably *more* overhead than just assigning it?
    logging.addLevelName(NOTIFY, "NOTIFY")
    logging.NOTIFY = NOTIFY
    logging.getLoggerClass().notify = notify


def set_root_logger(log_level=None, log_path=None):
    """Set the root logger 'pypyr'. Do this before you do anything else.

    Run once and only once at initialization.

    log_level enumeration:
    10=DEBUG
    20=INFO
    25=NOTIFY (default)
    30=WARNING
    40=ERROR
    50=CRITICAL

    < 10 gives full traceback on errors.

    Args:
        log_level (int): Log level. If not specified, defaults to 25 - NOTIFY.
        log_path (path-like): File path+name. If specified, pypyr will append
            logging output to this indefinitely growing file and to the
            console.
    """
    if config.log_config:
        logging.config.dictConfig(config.log_config)
    else:
        handlers = get_log_handlers(log_level, log_path)
        set_logging_config(log_level, handlers=handlers)

    root_logger = logging.getLogger('pypyr')
    root_logger.debug(
        "Root logger %s configured with level %s",
        root_logger.name, log_level)


def get_log_handlers(log_level, log_path=None):
    """Return list of log handlers to handle stdout, stderr and file."""
    handlers = []
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.addFilter(LevelFilter(logging.WARNING))

    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(max(logging.WARNING,
                                log_level or logging.NOTIFY))

    handlers.append(stdout_handler)
    handlers.append(stderr_handler)
    if log_path:
        file_handler = logging.FileHandler(log_path)
        handlers.append(file_handler)

    return handlers



"""switch.py unit tests. Most of the tests are in pypyr.steps.dsl.cof."""
import pytest

from pypyr.context import Context
from pypyr.errors import Call
from pypyr.steps.switch import run_step


def test_switch_step_calls_cof():
    """Switch step calls through to dsl."""
    with pytest.raises(Call) as err:
        run_step(Context({
            'switch': [
                {'case': False, 'call': 'sg1'},
                {'case': True, 'call': 'sg2'}
            ]
        }))

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['sg2']
    assert cof.success_group is None
    assert cof.failure_group is None
    assert cof.original_config == ('switch', [
        {'case': False, 'call': 'sg1'},
        {'case': True, 'call': 'sg2'},
    ])


def test_switch_step_nothing():
    """No success match on switch case just carries on."""
    run_step(Context({
        'switch': [
            {'case': False, 'call': 'sg1'},
            {'case': False, 'call': 'sg2'}
        ]
    }))



"""pypyr.steps.cmd unit tests.

The bulk of the unit tests are in:
tests/unit/steps/dsl/cmdasync_test.py

For tests involving stdout & stderr set to files, see:
tests/integration/pypyr/steps/cmds_int_test.py
"""
from unittest.mock import patch

from pypyr.context import Context
import pypyr.steps.cmds


def test_cmds_step():
    """Use the AsyncCmdStep handler to run the step."""
    with patch('pypyr.steps.cmds.AsyncCmdStep') as mock_runner:
        pypyr.steps.cmds.run_step(Context({'a': 'b'}))

    mock_runner.assert_called_once_with(name='pypyr.steps.cmds',
                                        context={'a': 'b'},
                                        is_shell=False)



from unittest.mock import call

# the entire point is to test imports when import package, even when not used.
import pypyr  # noqa: F401
from tests.common.utils import patch_logger


def test_init_package():
    """Import pypyr runs one off initialization code."""
    import logging
    with patch_logger(__name__, logging.NOTIFY) as mock_log:
        logging.getLogger(__name__).notify("arb")

    assert mock_log.mock_calls == [call('arb')]



"""debug.py unit tests."""
import datetime
import logging
from collections import OrderedDict
from unittest.mock import call, patch
import sys

from pypyr.context import Context
import pypyr.steps.debug as debug
from pypyr.dsl import Jsonify, PyString, SicString
from tests.common.utils import patch_logger

is_windows = sys.platform.startswith("win")


def test_no_inputs():
    """Dump entire context with no input specified."""
    context = Context({'k1': 'v1', 'k2': 'x{k1}x', 'k3': [0, 1, 2]})

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'k1': 'v1', 'k2': 'x{k1}x', 'k3': [0, 1, 2]}")]


def test_no_keys_with_formatting():
    """Dump entire context with no keys but with formatting."""
    context = Context({'k1': 'v1',
                       'k2': 'x{k1}x',
                       'k3': [0, 1, 2],
                       'debug': {'format': True}
                       })

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'debug': {'format': True}, 'k1': 'v1', "
             "'k2': 'xv1x', 'k3': [0, 1, 2]}")]


def test_key_str():
    """Dump single key without formatting."""
    context = Context({'k1': 'v1',
                       'k2': 'x{k1}x',
                       'k3': [0, 1, 2],
                       'debug': {'keys': 'k2'}
                       })

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'k2': 'x{k1}x'}")]


def test_key_str_format():
    """Dump single key with formatting."""
    context = Context({'k1': 'v1',
                       'k2': 'x{k1}x',
                       'k3': [0, 1, 2],
                       'debug': {'keys': 'k2', 'format': True}
                       })

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'k2': 'xv1x'}")]


def test_keys_list():
    """Dump list of keys without formatting."""
    context = Context({'k1': 'v1',
                       'k2': 'x{k1}x',
                       'k3': [0, 1, 2],
                       'debug': {'keys': ['k1', 'k2']}
                       })

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'k1': 'v1', 'k2': 'x{k1}x'}")]


def test_keys_list_format():
    """Dump list of keys with formatting."""
    context = Context({'k1': 'v1',
                       'k2': 'x{k1}x',
                       'k3': [0, 1, 2],
                       'debug': {'keys': ['k1', 'k2'], 'format': True}
                       })

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'k1': 'v1', 'k2': 'xv1x'}")]


def test_recursive():
    """Dump of the recursive objects must not fail."""
    arb_list = ['ham', 'eggs']
    arb_list.append(arb_list)

    context = Context({'k1': arb_list})

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n"
             "{'k1': ['ham', 'eggs',"
             f" <Recursion on list with id={id(arb_list)}>]}}")]


def test_key_sorting():
    """Dict keys represented alphabetically."""
    context = Context(
        {
            'k2': 'v2',
            'k1': {'z': 'vz', 'a': 'va'}
        }
    )

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n{'k1': {'a': 'va', 'z': 'vz'}, 'k2': 'v2'}")]


def test_multiline():
    """Check line breaks at large objects."""
    context = Context(
        {
            'key1': 'value1',
            'key2': 'value2',
            'key3': {
                'extra_large_key_name': 'with_extra_ordinary_key_value'
            },
        }
    )

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [
        call("\n"
             "{'key1': 'value1',\n"
             " 'key2': 'value2',\n"
             " 'key3': {'extra_large_key_name': "
             "'with_extra_ordinary_key_value'}}")]


def get_obj_hex(obj):
    """Get platform appropriate object hex."""
    if is_windows:
        return f'0x{id(obj):0{16}X}'
    else:
        return hex(id(obj))


def test_complex_object():
    """Check different complex objects, like error, date, type..."""
    py_str = PyString('py_arb')

    def arb_func():
        pass

    # autopep8: off
    # ignore flake warning to check lambda address
    arb_lambda = lambda x: x  # noqa: E731
    # autopep8: on

    context = Context({
        'bound_method': py_str.get_value,
        'date': datetime.date(2019, 10, 10),
        'datetime': datetime.datetime(2019, 10, 10),
        'exception': ValueError("Test", 'exc_arg'),
        'func': arb_func,
        'jsonify': Jsonify('jsonify arb'),
        'lambda': arb_lambda,
        'list': ['arb1', 'arb2', ['arb3']],
        'ordered_dict': OrderedDict({'a': 1, 'b': 2}),
        'py': py_str,
        'sic': SicString('sic_arb'),
        'tuple': ('arb1', 'arb2',),
        'type': ValueError,
    })

    with patch_logger('pypyr.steps.debug', logging.INFO) as mock_logger_info:
        debug.run_step(context)

    assert mock_logger_info.mock_calls == [call(
        "\n"
        "{'bound_method': <bound method PyString.get_value "
        "of PyString('py_arb')>,\n"
        " 'date': datetime.date(2019, 10, 10),\n"
        " 'datetime': datetime.datetime(2019, 10, 10, 0, 0),\n"
        " 'exception': ValueError('Test', 'exc_arg'),\n"
        " 'func': <function test_complex_object.<locals>.arb_func"
        f" at {get_obj_hex(arb_func)}>,\n"
        " 'jsonify': Jsonify('jsonify arb'),\n"
        " 'lambda': <function test_complex_object.<locals>.<lambda>"
        f" at {get_obj_hex(arb_lambda)}>,\n"
        " 'list': ['arb1', 'arb2', ['arb3']],\n"
        " 'ordered_dict': OrderedDict([('a', 1), ('b', 2)]),\n"
        " 'py': PyString('py_arb'),\n"
        " 'sic': SicString('sic_arb'),\n"
        " 'tuple': ('arb1', 'arb2'),\n"
        " 'type': <class 'ValueError'>}")]


@patch("pprint.pformat")
def test_pformat_called_when_logging_is_enabled(mock_pformat):
    """Check that pformat called only when logging is enabled."""
    ctx = Context({"k1": "v1"})
    debug.run_step(ctx)
    mock_pformat.assert_called_once_with(ctx)


@patch("pprint.pformat")
def test_pformat_not_called_when_logging_is_disabled(mock_pformat):
    """Check that pformat not called only when logging is disabled."""
    ctx = Context({"k1": "v1"})

    logging_level = debug.logger.level
    debug.logger.setLevel(logging.WARNING)
    try:
        debug.run_step(ctx)
    finally:
        debug.logger.setLevel(logging_level)

    mock_pformat.assert_not_called()



"""pypyr.steps.shells unit tests.

The bulk of the unit tests are in:
tests/unit/steps/dsl/cmdasync_test.py

For tests involving stdout & stderr set to files, see:
tests/integration/pypyr/steps/cmds_int_test.py
"""
from pathlib import Path

import pytest

from pypyr.config import config
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError, MultiError, SubprocessError
import pypyr.steps.shells

cmd_path = Path.cwd().joinpath('tests/testfiles/cmds')
is_windows = config.is_windows


def get_cmd(posix, win):
    """Return posix or win depending on current platform."""
    return win if is_windows else posix


def test_shells_str_with_args():
    """Single shell string works with string interpolation."""
    cmd = get_cmd('tests/testfiles/cmds/args.sh',
                  r'tests\testfiles\cmds\args.bat')
    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd,
                       'cmds': ['{d} {a} "{b}" {c}']})
    pypyr.steps.shells.run_step(context)

    assert 'cmdOut' not in context


def test_shells_sequence_with_ampersands_save_output():
    """Single shell command string with ampersands works."""
    context = Context({'one': 1,
                       'two': 2,
                       'three': 3,
                       'cmds': {
                           'run': ['echo {one} && echo {two} && echo {three}'],
                           'save': True}
                       })
    pypyr.steps.shells.run_step(context)

    out = context['cmdOut']
    assert len(out) == 1
    out1 = out[0]
    assert out1.returncode == 0
    assert out1.stdout == ('1 \r\n2 \r\n3' if is_windows
                           else '1\n2\n3')
    assert not out1.stderr


def test_shells_dict_input_with_string_interpolation_save_out():
    """Single command string works with string interpolation."""
    context = Context({'cmds': {'run': ["echo blah"],
                                'save': True}})
    pypyr.steps.shells.run_step(context)

    out = context['cmdOut']
    assert len(out) == 1
    out1 = out[0]

    assert out1.returncode == 0
    assert out1.stdout == 'blah'
    assert not out1.stderr


def test_shells_with_cwd():
    """Single command string works with string interpolation."""
    cmd = get_cmd('pwd', 'echo %cd%')
    context = Context({'cmds': {'run': cmd,
                                'save': True,
                                'cwd': 'tests'}})
    pypyr.steps.shells.run_step(context)

    out = context['cmdOut']
    assert len(out) == 1
    out1 = out[0]

    assert out1.returncode == 0
    assert Path(out1.stdout).samefile('tests')
    assert not out1.stderr


def test_shells_error_throws():
    """Shell process returning 1 should throw CalledProcessError."""
    with pytest.raises(MultiError) as the_err:
        context = Context({'cmds': 'exit 1'})
        pypyr.steps.shells.run_step(context)

    err = the_err.value
    assert len(err) == 1
    assert isinstance(err[0], SubprocessError)


def test_shells_empty_context_cmd_throw():
    """Empty cmds in context should throw assert error."""
    with pytest.raises(KeyNotInContextError) as err_info:
        context = Context({'blah': 'blah blah'})
        pypyr.steps.shells.run_step(context)

    assert str(err_info.value) == ("context['cmds'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.shells.")



"""Version information."""

import platform
from pypyr import __version__


def get_version():
    """Return package-name __version__ python python_version."""
    return (f'pypyr {__version__} '
            f'python {platform.python_version()}')



"""pipedef.py unit tests."""
from pypyr.pipedef import PipelineDefinition, PipelineInfo, PipelineFileInfo

# region PipelineDefinition


def test_pipelinedefinition_eq():
    """Check PipelineDefinition equality."""
    assert (PipelineDefinition(pipeline='pipe', info='info')
            == PipelineDefinition(pipeline='pipe', info='info'))
    assert (PipelineDefinition(pipeline='pipe', info='info')
            != PipelineDefinition(pipeline='pipe2', info='info2'))
    assert PipelineDefinition(pipeline='pipe', info='info') != 1

# endregion PipelineDefinition

# region PipelineInfo


def test_pipelineinfo_ctor():
    """Test PipelineInfo constructor defaults."""
    pi = PipelineInfo(pipeline_name='arbpipe',
                      parent='arbdir',
                      loader='arbloader')

    assert pi.pipeline_name == 'arbpipe'
    assert pi.parent == 'arbdir'
    assert pi.loader == 'arbloader'
    assert pi.is_parent_cascading
    assert pi.is_loader_cascading


def test_pipelineinfo_no_cascade():
    """Test PipelineInfo constructor no cascade."""
    pi = PipelineInfo(pipeline_name='arbpipe',
                      parent='arbdir',
                      loader='arbloader',
                      is_parent_cascading=False,
                      is_loader_cascading=False)

    assert pi.pipeline_name == 'arbpipe'
    assert pi.parent == 'arbdir'
    assert pi.loader == 'arbloader'
    assert not pi.is_parent_cascading
    assert not pi.is_loader_cascading


def test_pipelineinfo_eq():
    """Check PipelineInfo equality."""
    assert (PipelineInfo(pipeline_name='arbpipe',
                         parent='arbdir',
                         loader='arbloader')
            == PipelineInfo(pipeline_name='arbpipe',
                            parent='arbdir',
                            loader='arbloader'))

    assert (PipelineInfo(pipeline_name='arbpipe',
                         parent='arbdir',
                         loader='arbloader')
            != PipelineInfo(pipeline_name='arbpipe',
                            parent='arbdir2',
                            loader='arbloader'))

    assert PipelineInfo(pipeline_name='arbpipe',
                        parent='arbdir',
                        loader='arbloader') != 2
# endregion PipelineInfo

# region PipelineFileInfo


def test_pipelinefileinfo_ctor():
    """Test PipelineFileInfo constructor defaults."""
    pi = PipelineFileInfo(pipeline_name='arbpipe',
                          loader='arbloader',
                          parent='arbdir',
                          path='arbpath')

    assert pi.pipeline_name == 'arbpipe'
    assert pi.parent == 'arbdir'
    assert pi.loader == 'arbloader'
    assert pi.path == 'arbpath'
    assert pi.is_parent_cascading
    assert pi.is_loader_cascading


def test_pipelinefileinfo_eq():
    """Check PipelineInfo equality."""
    assert (PipelineFileInfo(pipeline_name='arbpipe',
                             loader='arbloader',
                             parent='arbdir',
                             path='arbpath')
            == PipelineFileInfo(pipeline_name='arbpipe',
                                loader='arbloader',
                                parent='arbdir',
                                path='arbpath'))

    assert (PipelineFileInfo(pipeline_name='arbpipe',
                             loader='arbloader',
                             parent='arbdir',
                             path='arbpath')
            != PipelineFileInfo(pipeline_name='arbpipe',
                                loader='arbloader',
                                parent='arbdir2',
                                path='arbpath'))

    assert (PipelineFileInfo(pipeline_name='arbpipe',
                             loader='arbloader',
                             parent='arbdir',
                             path='arbpath')
            != PipelineFileInfo(pipeline_name='arbpipe',
                                loader='arbloader',
                                parent='arbdir',
                                path='arbpath2'))

    assert (PipelineFileInfo(pipeline_name='arbpipe',
                             loader='arbloader',
                             parent='arbdir',
                             path='arbpath')
            != PipelineInfo(pipeline_name='arbpipe',
                            parent='arbdir',
                            loader='arbloader'))

    assert PipelineFileInfo(pipeline_name='arbpipe',
                            loader='arbloader',
                            parent='arbdir',
                            path='arbpath') != 2
# endregion PipelineFileInfo



"""pypyr step that imports to the namespace for PyString.

Make Python stdlib or your own code available to any PyString in the pipeline.

Only supports absolute imports, not relative.

Supports python import syntax like:
    import x
    import x as y

    import x.y
    import x.y as z

    from x import y
    from x import y as z

    from x import y, z
    from a.b import c as d, e as f
"""
import logging
from pypyr.cache.namespacecache import pystring_namespace_cache

logger = logging.getLogger(__name__)


def run_step(context):
    """Import modules, classes or functions for PyString expressions.

    Args:
        context (pypyr.context.Context):
            Context is a dictionary or dictionary-like.
            Context must contain key 'pyImport'
            The value of pyImport is a string containing python import
            statements.
    """
    logger.debug("started")
    context.assert_key_has_value(key='pyImport', caller=__name__)

    # block yaml style could be LiteralScalarString, which is not hashable for
    # cache key. Explicitly make a string.
    source = str(context['pyImport'])

    namespace = pystring_namespace_cache.get_namespace(source)
    # len safe coz get_namespace returns {}, not None.
    logger.debug("imported %s objects to merge into PyString namespace",
                 len(namespace))

    new_length = context.pystring_globals_update(namespace)

    logger.debug("PyString namespace now contains %s objects.", new_length)

    logger.debug("done")



"""filewritetoml.py unit tests."""
import io
from unittest.mock import mock_open, patch

import pytest

from pypyr.context import Context
from pypyr.errors import (
    ContextError,
    KeyInContextHasNoValueError,
    KeyNotInContextError)
import pypyr.steps.filewritetoml as filewrite


def test_filewritetoml_no_filewritetoml_raises():
    """No input fileWriteToml raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteToml'] doesn't exist. "
        "It must exist for pypyr.steps.filewritetoml.")


def test_filewritetoml_none_filewritetoml_raises():
    """None fileWriteToml raises."""
    context = Context({
        'k1': 'v1',
        'fileWriteToml': None})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteToml'] must have a value for "
        "pypyr.steps.filewritetoml.")


def test_filewritetoml_filewritetoml_not_iterable_raises():
    """Not iterable fileWriteToml raises."""
    context = Context({
        'k1': 'v1',
        'fileWriteToml': 1})

    with pytest.raises(ContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteToml'] must exist, be iterable "
        "and contain 'path' for pypyr.steps.filewritetoml. argument of type "
        "'int' is not iterable")


def test_filewritetoml_empty_path_raises():
    """Empty path raises."""
    context = Context({
        'fileWriteToml': {
            'path': None
        }})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == (
        "context['fileWriteToml']['path'] must have a value for "
        "pypyr.steps.filewritetoml.")


def test_filewritetoml_no_path_raises():
    """No path raises."""
    context = Context({
        'fileWriteToml': 'blah',
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filewrite.run_step(context)

    assert str(err_info.value) == ("context['fileWriteToml']['path'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.filewritetoml.")


@patch('pypyr.steps.filewritetoml.Path')
def test_filewritetoml_pass_no_payload(mock_path):
    """Success case writes all context out when no payload."""
    context = Context({
        'k1': 'v1',
        'fileWriteToml': {
            'path': '/arb/blah'
        }})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.toml.open', mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        output = out_bytes.getvalue().decode()

    mocked_path = mock_path.return_value
    mock_path.assert_called_once_with('/arb/blah')
    mock_output.assert_called_once_with(mocked_path, 'wb')
    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'v1'
    assert context['fileWriteToml'] == {'path': '/arb/blah'}

    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)

    # toml well formed & new lines and indents are where they should be
    assert output == 'k1 = "v1"\n\n[fileWriteToml]\npath = "/arb/blah"\n'


@patch('pypyr.steps.filewritetoml.Path')
def test_filewritetoml_pass_with_payload(mock_path):
    """Success case writes only specific context payload."""
    context = Context({
        'k1': 'v1',
        'fileWriteToml': {
            'path': '/arb/blah',
            'payload': {
                'one': 'v1',
                'two': True,
                'three': {'a': 'b', 'c': 123.45,
                          'd': [0, 1, 2]},
                'four': 12,
                'five': True
            }}})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.toml.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        output = out_bytes.getvalue().decode()

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'v1'

    assert context['fileWriteToml'] == {'path': '/arb/blah',
                                        'payload': {
                                            'one': 'v1',
                                            'two': True,
                                            'three': {'a': 'b', 'c': 123.45,
                                                      'd': [0, 1, 2]},
                                            'four': 12,
                                            'five': True
                                        }}

    mock_path.assert_called_once_with('/arb/blah')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'wb')

    # toml well formed & new lines + indents are where they should be
    assert output == ('one = "v1"\n'
                      'two = true\n'
                      'four = 12\n'
                      'five = true\n'
                      '\n'
                      '[three]\n'
                      'a = "b"\n'
                      'c = 123.45\n'
                      'd = [\n'
                      '    0,\n'
                      '    1,\n'
                      '    2,\n'
                      ']\n')


@patch('pypyr.steps.filewritetoml.Path')
def test_filewritetoml_pass_no_payload_substitutions(mock_path):
    """Success case writes all context out with substitutions."""
    context = Context({
        'k1': 'v1',
        'pathkey': '/arb/path',
        'parent': [0, 1, {'child': '{k1}'}],
        'nested': '{parent[2][child]}',
        'fileWriteToml': {
            'path': '{pathkey}'
        }})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.toml.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        output = out_bytes.getvalue().decode()

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['k1'] == 'v1'
    assert context['fileWriteToml'] == {'path': '{pathkey}'}

    mock_path.assert_called_once_with('/arb/path')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'wb')

    # toml well formed & new lines + indents are where they should be
    assert output == ('k1 = "v1"\n'
                      'pathkey = "/arb/path"\n'
                      'parent = [\n'
                      '    0,\n'
                      '    1,\n'
                      '    { child = "v1" },\n'
                      ']\n'
                      'nested = "v1"\n'
                      '\n'
                      '[fileWriteToml]\n'
                      'path = "/arb/path"\n')


@patch('pypyr.steps.filewritetoml.Path')
def test_filewritetoml_pass_with_payload_substitutions(mock_path):
    """Success case writes only specified context with substitutions."""
    context = Context({
        'k1': 'v1',
        'intkey': 3,
        'pathkey': '/arb/path',
        'parent': [0,
                   1,
                   {'child': ['{k1}',
                              '{intkey}',
                              ['a', 'b', 'c']
                              ]}],
        'nested': '{parent[2][child]}',
        'fileWriteToml': {
            'path': '{pathkey}',
            'payload': '{parent[2]}'
        }})

    with io.BytesIO() as out_bytes:
        with patch('pypyr.toml.open',
                   mock_open()) as mock_output:
            mock_output.return_value.write.side_effect = out_bytes.write
            filewrite.run_step(context)

        output = out_bytes.getvalue().decode()

    assert context, "context shouldn't be None"
    assert len(context) == 6, "context should have 6 items"
    assert context['k1'] == 'v1'
    assert context['fileWriteToml'] == {'path': '{pathkey}',
                                        'payload': '{parent[2]}'}
    assert context['parent'] == [0,
                                 1,
                                 {'child': ['{k1}',
                                            '{intkey}',
                                            ['a', 'b', 'c']
                                            ]
                                  }
                                 ]

    mock_path.assert_called_once_with('/arb/path')
    mocked_path = mock_path.return_value
    mocked_path.parent.mkdir.assert_called_once_with(parents=True,
                                                     exist_ok=True)
    mock_output.assert_called_once_with(mocked_path, 'wb')

    # toml well formed & new lines + indents are where they should be
    assert output == ('child = [\n'
                      '    "v1",\n'
                      '    3,\n'
                      '    [\n'
                      '        "a",\n'
                      '        "b",\n'
                      '        "c",\n'
                      '    ],\n'
                      ']\n')


def test_filewritetoml_fail_with_empty_payload():
    """Empty payload does not write empty file."""
    context = Context({
        'k1': 'v1',
        'fileWriteToml': {
            'path': '/arb/blah',
            'payload': ''
        }})

    with pytest.raises(KeyInContextHasNoValueError) as err:
        filewrite.run_step(context)

    assert str(err.value) == (
        'payload must have a value to write to output TOML document.')


def test_filewritetoml_fail_with_none_payload():
    """None payload does not write empty file."""
    context = Context({
        'k1': 'v1',
        'fileWriteToml': {
            'path': '/arb/blah',
            'payload': None
        }})

    with pytest.raises(KeyInContextHasNoValueError) as err:
        filewrite.run_step(context)

    assert str(err.value) == (
        'payload must have a value to write to output TOML document.')



"""keyvaluepairs.py unit tests."""
import pypyr.parser.keyvaluepairs


def test_kvp_args_parses_to_dict():
    """Comma delimited input kvp string should return dictionary."""
    out = pypyr.parser.keyvaluepairs.get_parsed_context(['key1=value1',
                                                         'key2=value2',
                                                         ',key3=value3'])
    assert out['key1'] == 'value1', "key1 should be value1."
    assert out['key2'] == 'value2', "key2 should be value2."
    assert out[',key3'] == 'value3', "key3 should be value3."
    assert len(out) == 3, "3 items expected"


def test_kvp_args_single_parses_to_single_entry():
    """No commas input kvp string should return dictionary with 1 item."""
    out = pypyr.parser.keyvaluepairs.get_parsed_context(['key 1=value 2 '])
    assert out['key 1'] == 'value 2 ', "key 1 isnt 'value 2 '."
    assert len(out) == 1, "1 item expected"


def test_kvp_args_single_no_equals():
    """No equals input kvp ends up as key: ''."""
    out = pypyr.parser.keyvaluepairs.get_parsed_context(
        ['key1value2,value3'])

    assert out == {'key1value2,value3': ''}


def test_kvp_args_no_equals_and_equals():
    """No equals input combined with equals key value pair."""
    out = pypyr.parser.keyvaluepairs.get_parsed_context(
        ['key1value2,value3', "key1=123"])

    assert out == {'key1value2,value3': '', 'key1': '123'}


def test_kvp_args_ignore_double_equals():
    """Ignore double equals."""
    out = pypyr.parser.keyvaluepairs.get_parsed_context(
        ['key1value2,value3', "key 1=123=45 6"])

    assert out == {'key1value2,value3': '', 'key 1': '123=45 6'}


def test_kvp_empty_args_empty_dict():
    """Empty input args should return empty dict."""
    out = pypyr.parser.keyvaluepairs.get_parsed_context(None)
    assert not out



"""pypyr step yaml definition for control of flow instructions."""
from __future__ import annotations
import logging
from collections.abc import Mapping

from pypyr.context import Context
from pypyr.errors import (Call,
                          ContextError,
                          ControlOfFlowInstruction,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)

SWITCH_KEY = 'switch'

sentinel = object()


def instruction_from_dict(config: Mapping,
                          name: str,
                          instruction_type: type[ControlOfFlowInstruction],
                          context_key: str,
                          original_config: tuple[str, Mapping]
                          ) -> ControlOfFlowInstruction:
    """Create a Call or Jump from input dict."""
    if isinstance(config, str):
        groups = [config]
        success_group = None
        failure_group = None
    elif isinstance(config, list):
        groups = config
        success_group = None
        failure_group = None
    elif isinstance(config, dict):
        if 'groups' not in config:
            raise KeyNotInContextError(
                f"{context_key} needs a child key 'groups', which should "
                "be a list or a str with the step-group name(s) you want "
                f"to run. This is for step {name}.")
        groups = config['groups']
        if not groups:
            raise KeyInContextHasNoValueError(
                f"{context_key}.groups must have a value for step {name}")

        if isinstance(groups, str):
            groups = [groups]

        success_group = config.get('success', None)
        failure_group = config.get('failure', None)
    else:
        raise ContextError(
            f"{context_key} needs a child key 'groups', which should "
            "be a list or a str with the step-group name(s) you want "
            f"to run. This is for step {name}. Instead, you've got {config}")

    if success_group is not None and not isinstance(success_group, str):
        raise ContextError(
            f"{context_key}.success must be a string for {name}.")

    if failure_group is not None and not isinstance(failure_group, str):
        raise ContextError(
            f"{context_key}.failure must be a string for {name}.")

    return instruction_type(groups=groups,
                            success_group=success_group,
                            failure_group=failure_group,
                            original_config=original_config)


def control_of_flow_instruction(
        name: str,
        instruction_type: type[ControlOfFlowInstruction],
        context: Context,
        context_key: str) -> None:
    """Run a control of flow instruction.

    The step config in the context dict looks like this:
        <<instruction-name>>: <<cmd string>>. Mandatory.

        OR, as a dict
        <<instruction-name:
            groups: <<str>> or <<list of str>> - mandatory.
            success: <<str>>
            failure: <<str>>

    Args:
        name: Unique name for step. Likely __name__ of calling step.
        instruction_type: Type - must inherit from
                          pypyr.errors.ControlOfFlowInstruction
        context: pypyr.context.Context. Look for config in this context
                 instance.
        context_key: str name of step config in context.

    """
    assert name, ("name parameter must exist for a ControlOfFlowStep.")
    assert context, ("context param must exist for ControlOfFlowStep.")
    # this way, logs output as the calling step, which makes more sense
    # to end-user than a mystery steps.dsl.blah logging output.
    logger = logging.getLogger(name)
    logger.debug("starting")

    context.assert_key_has_value(key=context_key, caller=name)
    original_config = (context_key, context[context_key])

    config = context.get_formatted(context_key)

    cof_instruction = instruction_from_dict(config=config,
                                            name=name,
                                            instruction_type=instruction_type,
                                            context_key=context_key,
                                            original_config=original_config)

    logger.info(
        ("step %s about to hand over control with %s: Will run groups: %s "
         " with success %s and failure %s"),
        name,
        context_key,
        cof_instruction.groups,
        cof_instruction.success_group,
        cof_instruction.failure_group)
    raise cof_instruction


def switch(context: Context, name: str) -> None:
    """Run a call instruction from switch where case applies.

    Context expects the following keys:
        switch:
            - case: expression. exec `call` when expression True.
              call: str. Single group-name to run.

            - default: exec `call` when none of the case expressions True.

    OR

        switch:
            - case: expression. exec call `when` expression True.
              call:
                groups: list[str] | str. Single group or list of groups to run.
                success: str. Name of group to run on success completion.
                failure: str. Name of group to run on something going wrong.

    `default` is optional. If you specify default, it must be the last item in
    the switch list.

    Args:
        context: pypyr.context.Context: input context
        name: name of calling step. Very like `pypyr.steps.switch`.
    """
    # this way, logs output as the calling step, which makes more sense
    # to end-user than a mystery steps.dsl.blah logging output.
    logger = logging.getLogger(name)
    logger.debug("starting")

    context.assert_key_has_value(key=SWITCH_KEY, caller=name)

    switch_config = context[SWITCH_KEY]
    # allows runtime to reset switch input when nested switches
    original_config = (SWITCH_KEY, switch_config)

    last_index = len(switch_config) - 1
    cof_instruction = None
    raw_call = None
    is_default = False
    case_expression = False

    for i, case in enumerate(switch_config):
        logger.debug("evaluating case %s", case)
        if i == last_index:
            # last one might be default
            default_case = case.get('default')

            # this is default on the last case in the list
            is_default = default_case is not None

        if is_default:
            logger.debug("no switch case True. running default...")
            raw_call = default_case
        else:
            # `case` could be None or ''.
            raw_expression = case.get('case', sentinel)

            if raw_expression is sentinel:
                raise KeyNotInContextError(
                    f"'case' not found in `switch` index {i}.")

            try:
                # `call` must exist
                raw_call = case['call']
            except KeyError as err:
                raise KeyNotInContextError(
                    f"'call' not found in `switch` index {i}.") from err

            # `call` must have a value
            if not raw_call:
                raise KeyInContextHasNoValueError(
                    f"'call' does not have a value at `switch` index {i}.")

            case_expression = context.get_formatted_as_type(raw_expression,
                                                            out_type=bool)
            if case_expression:
                logger.debug("case %s is True. running `call` group...",
                             raw_expression)

        if case_expression or is_default:
            # only doing expensive format if case True.
            cof_instruction = instruction_from_dict(
                context.get_formatted_value(raw_call),
                name=name,
                instruction_type=Call,
                context_key=SWITCH_KEY,
                original_config=original_config)
            break

    if cof_instruction:
        logger.info(
            ("step %s about to hand over control: Will call groups: %s "
             " with success %s and failure %s"),
            name,
            cof_instruction.groups,
            cof_instruction.success_group,
            cof_instruction.failure_group)
        raise cof_instruction
    else:
        logger.info("no matching case found in switch.")
        logger.debug("done")



"""Global cache for retry back-off callables.

Attributes:
    backoff_cache: Global instance of the retry back-off callable cache.
                   Use this attribute to access the cache from elsewhere.
"""
import logging
from pypyr.cache.cache import Cache
import pypyr.moduleloader
from pypyr.retries import builtin_backoffs

logger = logging.getLogger(__name__)


class BackoffCache(Cache):
    """Get retry strategy callable objects from the Backoff cache."""

    def __init__(self):
        """Initialize the cache with the built-in back-off strategies."""
        super().__init__()
        self._cache = builtin_backoffs.copy()

    def clear(self):
        """Clear the cache of all objects except built-in back-offs.."""
        with self._lock:
            # rather than iterating & selectively removing, just reset entirely
            self._cache = builtin_backoffs.copy()

    def get_backoff(self, name):
        """Get cached backoff callable. Adds to cache if not exist.

        Args:
            name: load the package.module.class specified & add the object
                reference to cache.

        Return:
            callable: the function/callable reference specified by name.
        """
        logger.debug("starting")

        backoff = self.get(name, lambda: load_backoff_callable(name))

        logger.debug("done")
        return backoff


# global instance of the cache. use this to access the cache from elsewhere.
backoff_cache = BackoffCache()


def load_backoff_callable(name):
    """Load the backoff retry callable specified by name.

    Args:
        name: (string) Absolute name of callable to load.

    Returns:
        Reference to the callable specified by name.
    """
    logger.debug("starting")

    module_name, dot, attr_name = name.rpartition('.')
    if not dot:
        # just a bare name, no dot, means must be in globals.
        backoff_callable = builtin_backoffs.get(name)

        if backoff_callable:
            logger.debug("found built-in backoff callable `%s`", name)
            return backoff_callable
        else:
            # note that built-in backoffs already in cache, so this will only
            # run for bare names that aren't built-in back-offs.
            raise ValueError(
                f"Trying to find back-off strategy '{name}'. If this is a "
                "built-in back-off strategy, are you sure you got the name "
                "right?\nIf you're trying to load a custom callable, name "
                "should be in format 'package.module.ClassName' or "
                "'mod.ClassName'.")

    backoff_module = pypyr.moduleloader.get_module(module_name)
    logger.debug("retry backoff strategy module loaded: %s", backoff_module)

    try:
        backoff_callable = getattr(backoff_module, attr_name)
    except AttributeError:
        logger.error(("Looking for retry back-off strategy %s. "
                      "The callable %s doesn't exist in module %s."),
                     name, attr_name, backoff_module)
        raise
    logger.debug("done")

    return backoff_callable



"""Pype integration tests. Pipelines in ./tests/pipelines/pype."""
from pypyr.pipelinerunner import run as pipeline_run
import tests.common.pipeline_runner as test_pipe_runner


def test_pype_pipearg_int():
    """Pype parses pypeArg string and accessible in child."""
    pipename = 'tests/pipelines/pype/pipeargs'

    # this one doesn't work well with checking context out because
    # it works with pipelines that don't inherit parent scope.
    # thus catch notify output instead.
    test_pipe_runner.assert_pipeline_notify_output_is(pipename, ['A',
                                                                 'B',
                                                                 'C',
                                                                 'C',
                                                                 'D'])


def test_pype_err_int():
    """Pype calls handles error with pipeline_name correctly."""
    pipename = 'tests/pipelines/pype/pipeerr'
    out = pipeline_run(pipename)

    assert out['out'] == ['A', 'B', 'C']


def test_pype_pyimport():
    """Pype while passing pyimports to child."""
    pipename = 'tests/pipelines/pype/pyimport/parent'
    out = pipeline_run(pipename)

    assert out['out'] == ['A', 'B', 'C', 'D']


def test_pype_child_dir_relative_to_parent():
    """Child pipeline & custom steps location relative to parent."""
    pipename = 'tests/pipelines/pype/relativenesting/rootpipe'
    out = pipeline_run(pipename)

    assert out['out'] == ['A', 'B', 'C', 'D']


def test_pype_relative_pipes():
    """Sequence of relative names to nested sub-dirs same pipe diff levels."""
    pipename = 'tests/pipelines/pype/relative-pipes/pipe-a'
    out = pipeline_run(pipename)

    assert out['out'] == ['A', 'B', 'C', 'D', 'D']



"""python.py unit tests."""
import sys

from pypyr.context import Context
import pypyr.steps.python


def test_python_executable():
    """Python executable writes into context."""
    context = Context()
    pypyr.steps.python.run_step(context)
    assert context['python'] == sys.executable
    assert len(context) == 1



"""fileloader.py unit tests."""
from pathlib import Path
from unittest.mock import call, mock_open, patch, PropertyMock

import pytest

import pypyr.cache.admin
from pypyr.errors import PipelineNotFoundError
import pypyr.loaders.file as fileloader
from pypyr.pipedef import PipelineDefinition, PipelineFileInfo

# region get_pipeline_path

cwd = Path.cwd()
cwd_tests = cwd.joinpath('tests')
pypyr_path = cwd.joinpath('pypyr', 'pipelines')


def test_get_pipeline_path_absolute_path():
    """Find a pipeline with absolute path."""
    abs_path = Path('tests/testpipelinewd.yaml').resolve()
    str_abs_sans_yaml = str(abs_path.with_suffix(''))

    path_found = fileloader.get_pipeline_path(str_abs_sans_yaml, None)

    expected_path = cwd_tests.joinpath('testpipelinewd.yaml')
    assert path_found == expected_path


def test_get_pipeline_path_absolute_path_ignore_parent():
    """Find a pipeline with absolute path ignore parent input."""
    abs_path = Path('tests/testpipelinewd.yaml').resolve()
    str_abs_sans_yaml = str(abs_path.with_suffix(''))

    path_found = fileloader.get_pipeline_path(str_abs_sans_yaml, 'parent')

    expected_path = cwd_tests.joinpath('testpipelinewd.yaml')
    assert path_found == expected_path


def test_get_pipeline_path_absolute_path_not_found():
    """Error when can't find pipeline with absolute path."""
    abs_path = Path('tests/XXXZZZ.yaml').resolve()
    str_abs_sans_yaml = str(abs_path.with_suffix(''))

    with pytest.raises(PipelineNotFoundError) as err:
        fileloader.get_pipeline_path(str_abs_sans_yaml, None)

    assert str(err.value) == f'{abs_path} does not exist.'


def test_get_pipeline_path_no_parent():
    """Find a pipeline in the working dir when no parent."""
    path_found = fileloader.get_pipeline_path(
        'tests/testpipelinewd', None)

    expected_path = cwd_tests.joinpath('testpipelinewd.yaml')

    assert path_found == expected_path


def test_get_pipeline_path_in_parent():
    """Find a pipeline in the parent dir pipelines."""
    path_found = fileloader.get_pipeline_path('testpipeline',
                                              'tests/pipelines')

    expected_path = cwd_tests.joinpath('pipelines',
                                       'testpipeline.yaml')

    assert path_found == expected_path


def test_get_pipeline_path_in_cwd_pipelines():
    """Find a pipeline in cwd/pipelines."""
    # artificially change the cwd constant
    # when py 3.10 can use () instead of \ for multiline "with"
    with patch('pypyr.config.Config.cwd',
               new_callable=PropertyMock(return_value=cwd_tests)), \
        patch('pypyr.loaders.file.cwd_pipelines_dir',
              new=cwd_tests.joinpath('pipelines')):
        path_found = fileloader.get_pipeline_path('testpipeline', None)

    assert path_found == cwd_tests.joinpath('pipelines', 'testpipeline.yaml')


def test_get_pipeline_path_in_pypyr_dir():
    """Find a pipeline in the pypyr install dir."""
    path_found = fileloader.get_pipeline_path('donothing',
                                              'tests')

    expected_path = cwd.joinpath('pypyr',
                                 'pipelines',
                                 'donothing.yaml')

    assert path_found == expected_path


def test_get_pipeline_path_raises_with_parent_doesnt_exist():
    """Raise PipelineNotFoundError no pipeline found, non-existing parent."""
    with pytest.raises(PipelineNotFoundError) as err:
        fileloader.get_pipeline_path('unlikelypipeherexyz',
                                     'unlikelyparentxyz')

    cwd_pipes_path = cwd.joinpath('pipelines')

    expected_msg = (
        f'unlikelypipeherexyz.yaml not found in any of the following:\n'
        f'{cwd}\n{cwd_pipes_path}\n{pypyr_path}')

    assert str(err.value) == expected_msg


def test_get_pipeline_path_raises_with_parent_exists():
    """Raise PipelineNotFoundError no pipeline found, existing parent."""
    with pytest.raises(PipelineNotFoundError) as err:
        fileloader.get_pipeline_path('unlikelypipeherexyz',
                                     'tests/pipelines')

    parent = cwd.joinpath('tests/pipelines')
    cwd_pipes_path = cwd.joinpath('pipelines')

    expected_msg = (
        f'unlikelypipeherexyz.yaml not found in any of the following:\n'
        f'{parent}\n{cwd}\n{cwd_pipes_path}\n{pypyr_path}')

    assert str(err.value) == expected_msg


def test_get_pipeline_path_raises_no_parent():
    """Failure to find pipeline raises PipelineNotFoundError, no parent."""
    with pytest.raises(PipelineNotFoundError) as err:
        fileloader.get_pipeline_path('unlikelypipeherexyz',
                                     None)

    cwd_pipes_path = cwd.joinpath('pipelines')

    expected_msg = (
        f'unlikelypipeherexyz.yaml not found in any of the following:\n'
        f'{cwd}\n{cwd_pipes_path}\n{pypyr_path}')

    assert str(err.value) == expected_msg


def test_get_pipeline_path_raises_parent_is_cwd():
    """Raise PipelineNotFoundError on not found when parent == cwd."""
    with pytest.raises(PipelineNotFoundError) as err:
        fileloader.get_pipeline_path('unlikelypipeherexyz',
                                     Path.cwd())

    cwd_pipes_path = cwd.joinpath('pipelines')

    expected_msg = (
        f'unlikelypipeherexyz.yaml not found in any of the following:\n'
        f'{cwd}\n{cwd_pipes_path}\n{pypyr_path}')

    assert str(err.value) == expected_msg

# endregion get_pipeline_path

# region get_pipeline_definition


@patch('pypyr.loaders.file.add_sys_path')
@patch('ruamel.yaml.YAML.load', return_value='mocked pipeline def')
@patch('pypyr.loaders.file.get_pipeline_path',
       return_value=Path('arb/path/x.yaml'))
def test_get_pipeline_definition_pass(mocked_get_path,
                                      mocked_yaml,
                                      mocked_add_sys):
    """get_pipeline_definition passes correct params to all methods."""
    fileloader.file_cache.clear()

    with patch('pypyr.loaders.file.open',
               mock_open(read_data='pipe contents')) as mocked_open:
        pipeline_def = fileloader.get_pipeline_definition(
            'pipename', '/parent/dir')

    assert pipeline_def == PipelineDefinition(
        'mocked pipeline def',
        PipelineFileInfo(pipeline_name='x.yaml',
                         loader='pypyr.loaders.file',
                         parent=Path('arb/path'),
                         path=Path('arb/path/x.yaml')))

    path = Path('arb/path/x.yaml')
    mocked_get_path.assert_called_once_with(pipeline_name='pipename',
                                            parent='/parent/dir')
    mocked_open.assert_called_once_with(path, encoding=None)
    mocked_yaml.assert_called_once_with(mocked_open.return_value)
    assert fileloader.file_cache._cache == {str(path): pipeline_def}
    mocked_add_sys.assert_called_once_with(Path('arb/path'))

    fileloader.file_cache.clear()


@patch('pypyr.loaders.file.add_sys_path')
@patch('ruamel.yaml.YAML.load', return_value='mocked pipeline def')
@patch('pypyr.loaders.file.get_pipeline_path',
       return_value=Path('arb/path/x.yaml'))
def test_get_pipeline_definition_clear_all_cache(mocked_get_path,
                                                 mocked_yaml,
                                                 mocked_add_sys):
    """Clear all cache clears the pipeline file cache."""
    fileloader.file_cache.clear()

    with patch('pypyr.loaders.file.open',
               mock_open(read_data='pipe contents')) as mocked_open:
        pipeline_def = fileloader.get_pipeline_definition(
            'pipename', '/parent/dir')

    assert pipeline_def == PipelineDefinition(
        'mocked pipeline def',
        PipelineFileInfo(pipeline_name='x.yaml',
                         loader='pypyr.loaders.file',
                         parent=Path('arb/path'),
                         path=Path('arb/path/x.yaml')))

    path = Path('arb/path/x.yaml')
    mocked_get_path.assert_called_once_with(pipeline_name='pipename',
                                            parent='/parent/dir')
    mocked_open.assert_called_once_with(path, encoding=None)
    mocked_yaml.assert_called_once_with(mocked_open.return_value)
    assert fileloader.file_cache._cache == {str(path): pipeline_def}
    mocked_add_sys.assert_called_once_with(Path('arb/path'))

    pypyr.cache.admin.clear_all()
    assert fileloader.file_cache._cache == {}


@patch('pypyr.loaders.file.add_sys_path')
@patch('ruamel.yaml.YAML.load', return_value='mocked pipeline def')
@patch('pypyr.loaders.file.get_pipeline_path',
       return_value=Path('arb/path/x.yaml'))
def test_get_pipeline_definition_from_cache(mocked_get_path,
                                            mocked_yaml,
                                            mocked_add_sys):
    """Known path returns from cache."""
    fileloader.file_cache.clear()

    with patch('pypyr.loaders.file.open',
               mock_open(read_data='pipe contents')) as mocked_open:
        pipeline_def_1 = fileloader.get_pipeline_definition(
            'pipename', '/parent/dir')
        pipeline_def_2 = fileloader.get_pipeline_definition(
            'pipename', '/parent/dir')

    assert pipeline_def_1 == pipeline_def_2 == PipelineDefinition(
        'mocked pipeline def',
        PipelineFileInfo(pipeline_name='x.yaml',
                         loader='pypyr.loaders.file',
                         parent=Path('arb/path'),
                         path=Path('arb/path/x.yaml')))

    # get path called X2, everything subsequent X1
    mocked_get_path.mock_calls == [
        call(pipeline_name='pipename', parent='/parent/dir'),
        call(pipeline_name='pipename', parent='/parent/dir')]

    path = Path('arb/path/x.yaml')
    mocked_open.assert_called_once_with(path, encoding=None)
    mocked_yaml.assert_called_once_with(mocked_open.return_value)
    mocked_add_sys.assert_called_once_with(Path('arb/path'))

    assert fileloader.file_cache._cache == {str(path): pipeline_def_1}
    assert fileloader.file_cache._cache[str(path)] is pipeline_def_1

    fileloader.file_cache.clear()


@patch('pypyr.loaders.file.add_sys_path')
@patch('ruamel.yaml.YAML.load', return_value='mocked pipeline def')
@patch('pypyr.loaders.file.get_pipeline_path',
       return_value=Path('arb/path/x.yaml'))
def test_get_pipeline_definition_with_encoding(mocked_get_path,
                                               mocked_yaml,
                                               mocked_add_sys):
    """Open pipeline with custom encoding."""
    fileloader.file_cache.clear()

    with patch('pypyr.loaders.file.open',
               mock_open(read_data='pipe contents')) as mocked_open:
        with patch('pypyr.config.config.default_encoding', 'arb'):
            pipeline_def = fileloader.get_pipeline_definition(
                'pipename', '/parent/dir')

    assert pipeline_def == PipelineDefinition(
        'mocked pipeline def',
        PipelineFileInfo(pipeline_name='x.yaml',
                         loader='pypyr.loaders.file',
                         parent=Path('arb/path'),
                         path=Path('arb/path/x.yaml')))

    mocked_get_path.assert_called_once_with(pipeline_name='pipename',
                                            parent='/parent/dir')
    path = Path('arb/path/x.yaml')
    mocked_open.assert_called_once_with(path, encoding='arb')
    assert fileloader.file_cache._cache == {str(path): pipeline_def}
    mocked_yaml.assert_called_once_with(mocked_open.return_value)
    mocked_add_sys.assert_called_once_with(Path('arb/path'))
    fileloader.file_cache.clear()


@patch('pypyr.loaders.file.add_sys_path')
@patch('pypyr.loaders.file.get_pipeline_path',
       return_value=Path('arb/path/x.yaml'))
def test_get_pipeline_definition_file_not_found(mocked_get_path,
                                                mocked_add_sys_path):
    """get_pipeline_definition raises file not found."""
    fileloader.file_cache.clear()

    with patch('pypyr.loaders.file.open',
               mock_open(read_data='pipe contents')) as mocked_open:
        mocked_open.side_effect = FileNotFoundError('deliberate err')
        with pytest.raises(FileNotFoundError):
            fileloader.get_pipeline_definition('pipename', '/arb/dir')

    mocked_get_path.assert_called_once_with(pipeline_name='pipename',
                                            parent='/arb/dir')

    assert fileloader.file_cache._cache == {}
    mocked_add_sys_path.assert_not_called()

    fileloader.file_cache.clear()
# endregion get_pipeline_definition



"""pypyr step that gets the full path to the current Python executable."""
import logging
import sys

logger = logging.getLogger(__name__)


def run_step(context):
    """Get the full path to the current Python executable.

    Args:
        Context is a dictionary or dictionary-like.
        Does not require any specific keys in context.
    """
    logger.debug("started")
    context['python'] = sys.executable
    logger.debug("done")
